from getpass import getpass

from fabric import api

DEFAULT_HOST = '127.0.0.1'


def dump(backup_path, database_name, user='root', host=None, password=None):
    """
    Backup MySQL database as a MySQL archive.
    If host is set to None, 127.0.0.1 will be used.
    If password is set to None, a prompt will ask a password.
    """
    if host is None:
        host = DEFAULT_HOST

    api.run('mysqldump {database_name} -h{host} -u{user} {password_param} > {backup_path}'
            .format(
                database_name=database_name,
                host=host,
                user=user,
                password_param=get_password_param(user, password),
                backup_path=backup_path,
            ))


def restore(backup_path, database_name, user='root', host=None, password=None):
    """
    Restore MySQL database.
    If host is set to None, 127.0.0.1 will be used.
    If password is set to None, a prompt will ask a password.
    """
    if host is None:
        host = DEFAULT_HOST

    api.run('mysql -h{host} -u{user} {password_param} {database_name} < {backup_path}'
            .format(
                database_name=database_name,
                host=host,
                user=user,
                password_param=get_password_param(user, password),
                backup_path=backup_path,
            ))


def get_password_param(user, password):
    """
    Ask a password in the prompt
    """
    if password is None:
        password = getpass('Enter database password for {user}: '
                           .format(user=user))

    return '-p{password}'.format(password=password) if password != '' else ''



"""
This module allows you to send/receive signals during your deployment process.

You can use it to create modules separated from your main fabfile and make them
respond to certain signals. For example consider the following example, in your
``fabfile.py``::

    # This will trigger a pre_deploy and post_deploy signal
    @signals.register
    def deploy():
        deploy_my_app()

In your separate module, use the ``on`` decorator to hook on this signal::

    # This goes in your separate module
    @signals.on('fabfile.pre_deploy')
    def my_hook():
        print("This is called at the beginning of the deploy task")

You can also emit signals independently::

    def deploy():
        signals.emit('pre_deploy_my_app')
        deploy_my_app()
        signals.emit('hurray_my_app_is_deployed')

The :py:func:`task` decorator wraps the default :py:func:`fabric.api.task`
decorator with the :py:func:`register` signal, allowing you to intercept
pre/post signals without the need of adding the :py:func:`register` decorator
or firing the signals yourself.
"""

from collections import defaultdict
from functools import wraps
import inspect
import logging

from fabric.api import task as fabric_task


logger = logging.getLogger(__name__)


_callbacks = defaultdict(list)


def emit(signal):
    """
    Make all the receivers of this signal aware that it's been sent.
    """
    logger.debug("Emit signal %s" % signal)

    for callback in _callbacks[signal]:
        logger.debug("Execute function %s from %s" % (callback.__name__, inspect.getfile(callback)))
        callback()


def register(function):
    """
    Decorator that will emit pre and post signals before and after the
    function is executed. The signals are named after the function so if your
    function is named ``seek_the_holy_grail()``, the signal
    ``pre_seek_the_holy_grail`` will be fired before your function gets
    executed and the ``post_seek_the_holy_grail`` signal will be fired after
    your function has been executed.
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        emit("{module}.pre_{function}".format(
            module=function.__module__,
            function=function.__name__
        ))

        return_value = function(*args, **kwargs)

        emit("{module}.post_{function}".format(
            module=function.__module__,
            function=function.__name__
        ))

        return return_value

    return wrapper


def on(signal):
    """
    Decorator that will call the given function when the given signal is
    emitted.
    """
    def wrapper(function):
        """
        Add the given callback to the list of receivers of the given signal.
        """
        _callbacks[signal].append(function)

    return wrapper


def task(function):
    """
    Convenience decorator that wraps the default Fabric task decorator with the
    :py:func:`register` decorator.
    """
    return fabric_task(register(function))



"""
Functions for Drupal sites.

The `drupal_root` environment variable is used in some functions of this module
to run the commands in the correct directory. Make sure it is set to the root
directory of your Drupal project.

This module requires `drush` to be installed on the remote server.
"""
from contextlib import nested

from fabric import api

from fabliip.file import file_exists


def drush(command):
    """
    Runs a drush command on the server.

    Requires the `drupal_root` environment variable to be set.
    """
    with api.cd(api.env.drupal_root):
        output = api.run('drush -y {command}'.format(command=command))

    return output


def enable_disable_modules(site=None):
    """
    Enables and disables modules on the Drupal install to reflect the status of
    the modules.enabled/disabled files.

    The optional `site` parameter allows you to have a multisite project
    with a global modules.enabled/disabled file and a site-specific
    modules.site.enabled/disabled file.
    """
    with nested(api.cd(api.env.project_root), api.hide('commands')):
        enabled_modules = set(api.run('cat modules.enabled').splitlines())
        disabled_modules = set(api.run('cat modules.disabled').splitlines())

        if site is not None:
            site_enabled_modules_file = 'modules.%s.enabled' % site
            site_disabled_modules_file = 'modules.%s.disabled' % site

            if (not file_exists(site_enabled_modules_file)
                    or not file_exists(site_disabled_modules_file)):
                raise Exception("Couldn't find the site-specific modules files"
                                " (modules.{site}.enabled and"
                                " modules.{site}.disabled)".format(site=site))

            site_enabled_modules = set(
                api.run('cat modules.%s.enabled' % site).splitlines()
            )
            site_disabled_modules = set(
                api.run('cat modules.%s.disabled' % site).splitlines()
            )

            enabled_modules |= site_enabled_modules
            disabled_modules |= site_disabled_modules

        enabled_modules -= disabled_modules

        current_enabled_modules = set(
            drush('pm-list --status=enabled --pipe').splitlines()
        )

        current_disabled_modules = set(
            drush('pm-list --status="disabled,not installed" --pipe')
            .splitlines()
        )

    modules_to_enable = current_disabled_modules & enabled_modules
    if modules_to_enable:
        print("The following modules are being enabled: {modules}".format(
            modules=", ".join(modules_to_enable))
        )
    else:
        print("No modules to enable")

    for module in modules_to_enable:
        drush('pm-enable %s' % module)
        clear_cache()

    modules_to_disable = current_enabled_modules & disabled_modules
    if modules_to_disable:
        print("The following modules are being disabled: {modules}".format(
            modules=", ".join(modules_to_disable))
        )
    else:
        print("No modules to disable")

    for module in modules_to_disable:
        drush('pm-disable %s' % module)
        clear_cache()


def set_maintenance_mode(enabled):
    """
    Enables or disables the maintenance mode.
    """
    drush('vset maintenance_mode %s' % ('1' if enabled else '0'))


def clear_cache():
    """
    Clears the Drupal cache.
    """
    drush('cc all')



from getpass import getpass

from fabric import api


def dump(backup_path, database_name, user='postgres', host=None,
         password=None):
    """
    Backs up the given database to the given file as a PostgreSQL archive. If
    host is set to None, a local connection will be used, so you'll need to be
    able to sudo to the given user. Otherwise, a standard password connection
    will be used and the user will be asked for a password.
    """
    if host is None:
        api.sudo('pg_dump -Fc {database_name} > {backup_path}'
                 .format(
                     database_name=database_name,
                     backup_path=backup_path,
                 ), user=user)
    else:
        if password is None:
            password = getpass('Enter database password for {user}: '
                                  .format(user=user))

        with api.shell_env(PGPASSWORD=password):
            api.run('pg_dump -Fc -U {user} -h {host} {database_name} >'
                    ' {backup_path}'
                    .format(
                        user=user,
                        host=host,
                        database_name=database_name,
                        backup_path=backup_path,
                    ))
